import importlib
import ntpath
import sys
import inspect
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List
from hedra.logging import HedraLogger
from hedra.core.graphs.events.event import Event, EventHook
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.common.plugin import Plugin
from .stage import Stage


def import_from_path(path: str) -> Any:
    package_dir = Path(path).resolve().parent
    package_dir_path = str(package_dir)
    package_dir_module = package_dir_path.split('/')[-1]
    
    package = ntpath.basename(path)
    package_slug = package.split('.')[0]
    spec = importlib.util.spec_from_file_location(f'{package_dir_module}.{package_slug}', path)
    
    if path not in sys.path:
        sys.path.append(str(package_dir.parent))

    module = importlib.util.module_from_spec(spec)

    sys.modules[module.__name__] = module

    spec.loader.exec_module(module)

    return module


def import_stages(path: str) -> Dict[str, Stage]:
    
    module = import_from_path(path)
    direct_decendants = list({cls.__name__: cls for cls in Stage.__subclasses__()}.values())

    discovered = {}
    for name, stage_candidate in inspect.getmembers(module):
        if inspect.isclass(stage_candidate) and issubclass(stage_candidate, Stage) and stage_candidate not in direct_decendants:
            discovered[name] = stage_candidate

    return discovered


def import_plugins(path: str) -> Dict[PluginType, Dict[str, Plugin]]:
    module = import_from_path(path)
    direct_decendants = list({cls.__name__: cls for cls in Plugin.__subclasses__()}.values())

    plugins_by_type = defaultdict(dict)
    for name, plugin_candidate in inspect.getmembers(module):
        if inspect.isclass(plugin_candidate) and issubclass(plugin_candidate, Plugin) and plugin_candidate not in direct_decendants:
            plugins_by_type[plugin_candidate.type][plugin_candidate.name] = plugin_candidate

    return plugins_by_type


def set_stage_hooks(stage: Stage) -> Stage:
    methods = inspect.getmembers(stage, predicate=inspect.ismethod) 

    for _, method in methods:
        method_name = method.__qualname__
        hook: Hook = registrar.all.get(method_name)
        
        if hook:
            hook._call = hook._call.__get__(stage, stage.__class__)
            setattr(stage, hook.shortname, hook._call)

            if inspect.ismethod(hook.call) is False:
                hook.call = hook.call.__get__(stage, stage.__class__)
                setattr(stage, hook.shortname, hook.call)

            hook.stage = stage.name
            hook.stage_instance: Stage = stage
            
            stage.hooks[hook.hook_type].append(hook)

    return stage


def set_events(event_hooks: List[EventHook], logging: HedraLogger, metadata_string: str) -> List[EventHook]:

    for event_hook in event_hooks:
        for target_hook_name in event_hook.names:    
            target_hook = registrar.all.get(target_hook_name)

            if target_hook and target_hook.stage_instance:
                logging.filesystem.sync['hedra.core'].info(
                    f'{metadata_string} - Appendng Event - {event_hook.name}:{event_hook.hook_id} - to target Stage - {target_hook.stage}:{target_hook.stage_instance.stage_id} Event Hooks'
                )

                event = Event(target_hook, event_hook)
                event.target_key = event_hook.key

                target_hook_idx = -1

                try:
                    target_hook_names = [hook.name for hook in target_hook.stage_instance.hooks[target_hook.hook_type]]
                    target_hook_idx = target_hook_names.index(target_hook.name)

                except ValueError:
                    pass

                if target_hook_idx >= 0 and isinstance(target_hook, Event):
                    if event_hook.pre is True:
                        target_hook.pre_sources[event_hook.name] = event_hook
                        target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = target_hook

                    else:
                        target_hook.post_sources[event_hook.name] = event_hook
                        target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = target_hook
                    
                    registrar.all[event.name] = target_hook

                elif target_hook_idx >= 0:
                    target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = event
                    registrar.all[event.name] = event
                
                target_hook.stage_instance.linked_events[(target_hook.stage, target_hook.hook_type, target_hook.name)].append(
                    (event_hook.stage, event_hook.hook_type, event_hook.name)
                )

    return event_hooks


from typing import Any, Dict
from hedra.core.graphs.hooks.registry.registry_types import EventHook
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook


class Event:

    def __init__(self, target: Hook, source: EventHook) -> None:

        target.is_event = True

        self.target = target
        self.pre_sources: Dict[str, EventHook] = {}
        self.post_sources: Dict[str, EventHook] = {}
        self.target_name = None
        self.target_shortname = None
        self.target_key = None

        if source.pre:
            self.pre_sources[source.name] = source

        else:
            self.post_sources[source.name] = source

        if target:
            self.target_name = self.target.name
            self.target_shortname = self.target.shortname

        self.pre = source.pre
        self.as_hook = False

    def __getattribute__(self, name: str) -> Any:
        
        target = object.__getattribute__(self, 'target')

        if target and hasattr(target, name) and name != 'call':
            return getattr(target, name)
        
        return object.__getattribute__(self, name)

    async def call(self, *args, **kwargs):

        result = None

        if self.target:
            
            result = await self.execute_pre()
            result = await self.target.call(*args, **kwargs) 
            result = await self.execute_post(result)

        else:

            for source in self.pre_sources.values():
                result = await source.call()

                if source.key:
                    source.stage_instance.context[source.key] = result


            for source in self.post_sources.values():
                result = await source.call()

                if source.key:
                    source.stage_instance.context[source.key] = result

        return result

    async def execute_pre(self):

        result = None

        for source in self.pre_sources.values():
            result = await source.call()

            if source.key:
                source.stage_instance.context[source.key] = result
                self.target.stage_instance.context[source.key] = result

        return result

    async def execute_post(self, result: Any):
        for source in self.post_sources.values():
            source_hook_input = result
            if source.key:
                source_hook_input = self.target.stage_instance.context[source.key]

                if source_hook_input is None:
                    source_hook_input = source.stage_instance.context[source.key]

                if source_hook_input is None:
                    source_hook_input = result

            source_result = await source.call(source_hook_input)

            if source.key:
                source.stage_instance.context[source.key] = source_result
                self.target.stage_instance.context[source.key] = source_result

            else:
                result = source_result
        
        return result


from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.events.event import Event
from hedra.core.graphs.hooks.registry.registry_types import MetricHook
from hedra.core.graphs.stages.validate.exceptions import HookValidationError
from .base_hook_validator import BaseHookVaidator


class MetricHookVaidator(BaseHookVaidator):

    def __init__(self, metadata_string: str) -> None:
        super(
            MetricHookVaidator,
            self
        ).__init__(metadata_string)

    async def validate(self, hook: MetricHook):

        try:
            call = hook._call
            if isinstance(hook, Event):
                call = hook.target._call
                
            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Validating {hook.hook_type.name.capitalize()} Hook - {hook.name}:{hook.hook_id}:{hook.hook_id}')

            assert hook.hook_type is HookType.METRIC, f"Hook type mismatch - hook {hook.name}:{hook.hook_id} is a {hook.hook_type.name} hook, but Hedra expected a {HookType.METRIC.name} hook."
            assert hook.shortname in hook.name, f"Shortname {hook.shortname} must be contained in full Hook name {hook.name}:{hook.hook_id} for @metric hook {hook.name}:{hook.hook_id}."
            assert call is not None, f"Method is not not found on stage or was not supplied to hook - {hook.name}:{hook.hook_id}"
            assert call.__code__.co_argcount > 1, f"Missing required argument 'events_group' for @metric hook {hook.name}:{hook.hook_id}"
            assert call.__code__.co_argcount < 3, f"Too many args. - @metric hook {hook.name}:{hook.hook_id} only requires 'events_group' as additional arg."
            assert 'self' in call.__code__.co_varnames

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Validated {hook.hook_type.name.capitalize()} Hook - {hook.name}:{hook.hook_id}:{hook.hook_id}')

        except AssertionError as hook_validation_error:
                raise HookValidationError(hook.stage_instance, str(hook_validation_error))


from pydantic import BaseModel, Field, StrictStr


class RestoreHookValidator(BaseModel):
    key: StrictStr=Field(..., min_length=1)
    restore_filepath: StrictStr=Field(..., min_length=1)


class RestoreValidator:

    def __init__(__pydantic_self__, key: str, restore_filepath: str) -> None:
        RestoreHookValidator(
            key=key,
            restore_filepath=restore_filepath
        )


from typing import Any, Tuple, Dict
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.engines.types.task.result import TaskResult
from .base_event import BaseEvent


class TaskEvent(BaseEvent):

    def __init__(
        self, 
        stage: Any, 
        result: TaskResult
    ) -> None:
        super().__init__(
            stage,
            result
        )

        self.time = result.complete - result.start
        self.timings = {
            'total': self.time,
            'waiting': result.start - result.wait_start,
            'reading': result.complete - result.write_end
        }

        self.data = result.data




import asyncio
import time
import dill
import json
import threading
import os
import signal
import traceback
from collections import defaultdict
from typing import Any, Dict, List, Tuple
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.core.graphs.events.event import Event, EventHook
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.core.graphs.stages.base.import_tools import (
    import_stages,
    set_stage_hooks
)
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)
from hedra.reporting.events import EventsGroup


async def process_batch(
        stage: Stage,
        stage_name: str, 
        custom_metric_hook_names: List[str],
        results_batch: List[BaseResult],
        metadata_string: str
    ):

        hedra_config_filepath = os.path.join(
            os.getcwd(),
            '.hedra.json'
        )

        hedra_config = {}
        if os.path.exists(hedra_config_filepath):
            with open(hedra_config_filepath, 'r') as hedra_config_file:
                hedra_config = json.load(hedra_config_file)

        logging_config = hedra_config.get('logging', {})
        logfiles_directory = logging_config.get(
            'logfiles_directory',
            os.getcwd()
        )

        log_level = logging_config.get('log_level', 'info')

        logging_manager.disable(
            LoggerTypes.DISTRIBUTED,
            LoggerTypes.DISTRIBUTED_FILESYSTEM
        )

        logging_manager.update_log_level(log_level)
        logging_manager.logfiles_directory = logfiles_directory


        logger = HedraLogger()
        logger.initialize()
        await logger.filesystem.aio.create_logfile('hedra.core.log')
        await logger.filesystem.aio.create_logfile('hedra.reporting.log')

        logger.filesystem.create_filelogger('hedra.core.log')
        logger.filesystem.create_filelogger('hedra.reporting.log')

        await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Initializing results aggregation')

        start = time.monotonic()

        events =  defaultdict(EventsGroup)

        custom_metric_hooks = []
        for metric_hook_name in custom_metric_hook_names:
            custom_metric_hook = registrar.all.get(metric_hook_name)
            custom_metric_hooks.append(custom_metric_hook)

        await asyncio.gather(*[
            asyncio.create_task(events[stage_result.name].add(
                    stage,
                    stage_result,
                    stage_name
                )) for stage_result in results_batch
        ])
        
        for events_stage_name, events_group in events.items():  

            await logger.filesystem.aio['hedra.reporting'].debug(
                f'{metadata_string} - Events group - {events_group.events_group_id} - created for Stage - {events_stage_name} - Processed - {events_group.total}'
            )

            events_group.calculate_partial_group_stats()

        elapsed = time.monotonic() - start

        await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Results aggregation complete - Took: {round(elapsed, 2)} seconds')

        return dill.dumps(events)


def process_results_batch(config: Dict[str, Any]):
    import asyncio
    import uvloop
    uvloop.install()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    
    graph_name = config.get('graph_name')
    graph_path = config.get('graph_path')
    graph_id = config.get('graph_id')
    source_stage_name = config.get('source_stage_name')
    source_stage_context = config.get('source_stage_context')
    source_stage_id = config.get('source_stage_id')

    thread_id = threading.current_thread().ident
    process_id = os.getpid()

    meetadata_string = f'Graph - {graph_name}:{graph_id} - thread:{thread_id} - process:{process_id} - Stage: {source_stage_name}:{source_stage_id} - '

    stage_name = config.get('analyze_stage_name')
    custom_metric_hook_names = config.get('analyze_stage_metric_hooks', [])
    results_batch = config.get('analyze_stage_batched_results', [])
    analyze_stage_linked_events = config.get('analyze_stage_linked_events', defaultdict(list))

    discovered = import_stages(graph_path)
    stage: Stage = discovered.get(stage_name)()
    setup_analyze_stage: Stage = set_stage_hooks(stage)
    setup_analyze_stage.context.update(source_stage_context)

    loaded_stages: Dict[str, Stage] = {
        setup_analyze_stage.name: setup_analyze_stage
    }
    pipeline_stage = {
        setup_analyze_stage.name: setup_analyze_stage
    }

    for event_target in analyze_stage_linked_events:
        target_hook_stage, target_hook_type, target_hook_name = event_target
        for event_source in analyze_stage_linked_events[event_target]:
            event_source_stage, _, event_hook_name = event_source

            if target_hook_stage == stage_name:

                source_stage: Stage = loaded_stages.get(stage_name)
                source_stage: Stage = loaded_stages.get(event_source_stage)

                if source_stage is None:
                    source_stage = discovered.get(event_source_stage)()
                    source_stage = set_stage_hooks(source_stage)
                    
                    loaded_stages[source_stage.name] = source_stage


                pipeline_stage[source_stage.name] = source_stage

                event_hooks = [hook.name for hook in source_stage.hooks[HookType.EVENT]]
                event_hook_idx = event_hooks.index(event_hook_name)

                event_hook: EventHook = source_stage.hooks[HookType.EVENT][event_hook_idx]

                target_hook_names = [hook.name for hook in setup_analyze_stage.hooks[target_hook_type]]
                target_hook_idx = target_hook_names.index(target_hook_name)

                target_hook: Hook = setup_analyze_stage.hooks[target_hook_type][target_hook_idx]
                target_hook.stage = source_stage_name
                target_hook.stage_instance = setup_analyze_stage
                target_hook.stage_instance.hooks = setup_analyze_stage.hooks

                event = Event(target_hook, event_hook)
                event.target_key = event_hook.key

                if target_hook_idx >= 0 and isinstance(target_hook, Event):
                    if event_hook.pre is True:
                        target_hook.pre_sources[event_hook.name] = event_hook
                        target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = target_hook

                    else:
                        target_hook.post_sources[event_hook.name] = event_hook
                        target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = target_hook
                    
                    registrar.all[event.name] = target_hook

                elif target_hook_idx >= 0:
                    target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = event
                    registrar.all[event.name] = event
                
                target_hook.stage_instance.linked_events[(target_hook.stage, target_hook.hook_type, target_hook.name)].append(
                    (event_hook.stage, event_hook.hook_type, event_hook.name)
                )
                
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    def handle_loop_stop(signame):
        try:
            loop.close()
        except RuntimeError:
            os._exit(1)

    for signame in ('SIGINT', 'SIGTERM'):
        loop.add_signal_handler(
            getattr(signal, signame),
            lambda signame=signame: handle_loop_stop(signame)
        )

    try:
        events = loop.run_until_complete(
            process_batch(
                setup_analyze_stage,
                stage_name,
                custom_metric_hook_names,
                results_batch,
                meetadata_string
            )
        )

        context = {}
        for stage in pipeline_stage.values():
            context.update({
                context_key: context_value for context_key, context_value in stage.context.items() if context_key not in stage.context.known_keys
            })

        return {
            'events': events,
            'context': context
        }

    except BrokenPipeError:

        try:
            loop.close()
        except RuntimeError:
            pass

        return b''

    except Exception as e:
        raise e



import json
import dill
import threading
import os
import traceback
import signal
import os
from typing import Any, Dict, List, Union, Tuple
from hedra.core.engines.client.config import Config
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.graphs.stages.optimize.optimization import Optimizer
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.setup.setup import Setup
from hedra.core.graphs.events.event import Event, EventHook
from hedra.core.engines.types.registry import registered_engines
from hedra.core.personas.persona_registry import registered_personas
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.engine.engine_plugin import EnginePlugin
from hedra.plugins.types.persona.persona_plugin import PersonaPlugin
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.core.graphs.stages.execute import Execute
from hedra.core.graphs.stages.base.import_tools import (
    import_stages, 
    import_plugins,
    set_stage_hooks
)
from hedra.core.graphs.stages.optimize.optimization.algorithms import registered_algorithms
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)
from hedra.core.personas import get_persona




def optimize_stage(serialized_config: str):
    import asyncio
    import uvloop
    uvloop.install()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:

        hedra_config_filepath = os.path.join(
            os.getcwd(),
            '.hedra.json'
        )

        hedra_config = {}
        if os.path.exists(hedra_config_filepath):
            with open(hedra_config_filepath, 'r') as hedra_config_file:
                hedra_config = json.load(hedra_config_file)

        logging_config = hedra_config.get('logging', {})
        logfiles_directory = logging_config.get(
            'logfiles_directory',
            os.getcwd()
        )

        log_level = logging_config.get('log_level', 'info')

        logging_manager.disable(
            LoggerTypes.DISTRIBUTED,
            LoggerTypes.DISTRIBUTED_FILESYSTEM
        )

        logging_manager.update_log_level(log_level)
        logging_manager.logfiles_directory = logfiles_directory


        logger = HedraLogger()
        logger.initialize()
        logger.filesystem.sync.create_logfile('hedra.core.log')
        logger.filesystem.sync.create_logfile('hedra.optimize.log')

        logger.filesystem.create_filelogger('hedra.core.log')
        logger.filesystem.create_filelogger('hedra.optimize.log')

        thread_id = threading.current_thread().ident
        process_id = os.getpid()

        optimization_config: Dict[str, Union[str, int, Any]] = dill.loads(serialized_config)

        graph_name: str = optimization_config.get('graph_name')
        graph_path: str= optimization_config.get('graph_path')
        graph_id: str = optimization_config.get('graph_id')
        source_stage_name: str = optimization_config.get('source_stage_name')
        source_stage_id: str = optimization_config.get('source_stage_id')
        source_stage_context: Dict[str, Any] = optimization_config.get('source_stage_context')
        execute_stage_name: str = optimization_config.get('execute_stage_name')
        execute_stage_config: Config = optimization_config.get('execute_stage_config')
        execute_stage_plugins: Dict[PluginType, List[str]] = optimization_config.get('execute_stage_plugins')
        execute_stage_linked_events: Dict[Tuple[HookType, str], List[Tuple[str, str]]] = optimization_config.get('execute_stage_linked_events')
        optimize_params: List[str] = optimization_config.get('optimize_params')
        optimize_iterations: int = optimization_config.get('optimizer_iterations')
        optimizer_algorithm: str = optimization_config.get('optimizer_algorithm')
        time_limit: int = optimization_config.get('time_limit')
        batch_size: int = optimization_config.get('execute_stage_batch_size')

        metadata_string = f'Graph - {graph_name}:{graph_id} - thread:{thread_id} - process:{process_id} - Stage: {source_stage_name}:{source_stage_id} - '

        discovered = import_stages(graph_path)

        execute_stage: Stage = discovered.get(execute_stage_name)()
        execute_stage.context.update(source_stage_context)
        execute_stage = set_stage_hooks(execute_stage)

        execute_stage_config.batch_size = batch_size
        plugins_by_type = import_plugins(graph_path)

        stage_persona_plugins: List[str] = execute_stage_plugins[PluginType.PERSONA]
        persona_plugins: Dict[str, PersonaPlugin] = plugins_by_type[PluginType.PERSONA]

        stage_engine_plugins: List[str] = execute_stage_plugins[PluginType.ENGINE]
        engine_plugins: Dict[str, EnginePlugin] = plugins_by_type[PluginType.ENGINE]
        
        for plugin_name in stage_persona_plugins:
            plugin = persona_plugins.get(plugin_name)
            plugin.name = plugin_name
            registered_personas[plugin_name] = lambda config: plugin(config)
        
        for plugin_name in stage_engine_plugins:
            plugin = engine_plugins.get(plugin_name)
            plugin.name = plugin_name
            registered_engines[plugin_name] = lambda config: plugin(config)

        for plugin_name, plugin in plugins_by_type[PluginType.OPTIMIZER].items():
            registered_algorithms[plugin_name] = plugin

        setup_stage = Setup()
        setup_stage.logger.spinner.logger.log_level = 'critical'
        setup_stage.plugins_by_type = plugins_by_type
        setup_stage.generation_setup_candidates = 1
        setup_stage.stages[execute_stage.name] = execute_stage
        setup_stage.config = execute_stage_config

        stages = loop.run_until_complete(setup_stage.run())
        
        setup_execute_stage: Execute = stages.get(execute_stage_name)

        setup_execute_stage_hooks = {}
        for hook_type in setup_execute_stage.hooks:
            setup_execute_stage_hooks.update({
                hook.name: hook for hook in setup_execute_stage.hooks[hook_type]
            })

        loaded_stages: Dict[str, Stage] = {
            setup_execute_stage.name: setup_execute_stage
        }
        pipeline_stages = {
            setup_execute_stage.name: setup_execute_stage
        }
        
        for event_target in execute_stage_linked_events:
            target_hook_stage, target_hook_type, target_hook_name = event_target
            for event_source in execute_stage_linked_events[event_target]:
                event_source_stage, event_hook_type, event_hook_name = event_source
                
                if target_hook_stage == setup_execute_stage.name:

                    source_stage: Stage = loaded_stages.get(event_source_stage)
                    if source_stage is None:
                        source_stage = discovered.get(event_source_stage)()
                        source_stage = set_stage_hooks(source_stage)
                        
                        loaded_stages[source_stage.name] = source_stage

                    pipeline_stages[source_stage.name] = source_stage

                    event_hooks = [hook.name for hook in source_stage.hooks[HookType.EVENT]]
                    event_hook_idx = event_hooks.index(event_hook_name)

                    event_hook: EventHook = source_stage.hooks[HookType.EVENT][event_hook_idx]

                    target_hook_names = [hook.name for hook in setup_execute_stage.hooks[target_hook_type]]
                    target_hook_idx = target_hook_names.index(target_hook_name)

                    target_hook: Hook = setup_execute_stage.hooks[target_hook_type][target_hook_idx]

                    event = Event(target_hook, event_hook)
                    event.target_key = event_hook.key

                    if target_hook_idx >= 0 and isinstance(target_hook, Event):
                        if event_hook.pre is True:
                            target_hook.pre_sources[event_hook.name] = event_hook
                            target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = target_hook

                        else:
                            target_hook.post_sources[event_hook.name] = event_hook
                            target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = target_hook
                        
                        registrar.all[event.name] = target_hook

                    elif target_hook_idx >= 0:
                        target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = event
                        registrar.all[event.name] = event
                    
                    target_hook.stage_instance.linked_events[(target_hook.stage, target_hook.hook_type, target_hook.name)].append(
                        (event_hook.stage, event_hook.hook_type, event_hook.name)
                    )

        logger.filesystem.sync['hedra.optimize'].info(f'{metadata_string} - Setting up Optimization')

        optimizer = Optimizer({
            'graph_name': graph_name,
            'graph_id': graph_id,
            'source_stage_name': source_stage_name,
            'source_stage_id': source_stage_id,
            'params': optimize_params,
            'stage_name': execute_stage_name,
            'stage_config': execute_stage_config,
            'stage_hooks': setup_execute_stage.hooks,
            'iterations': optimize_iterations,
            'algorithm': optimizer_algorithm,
            'time_limit': time_limit
        })

        def handle_loop_stop(signame):
            try:
                optimizer._event_loop.stop()
                optimizer._event_loop.close()

            except BrokenPipeError:
                os._exit(1)
                
            except RuntimeError:
                os._exit(1)

        for signame in ('SIGINT', 'SIGTERM'):
            loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(signame)
            )

            optimizer._event_loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(signame)
            )

        results = optimizer.optimize()

        execute_stage_config.batch_size = results.get('optimized_batch_size', execute_stage_config.batch_size)
        execute_stage_config.batch_interval = results.get('optimized_batch_interval', execute_stage_config.batch_gradient)
        execute_stage_config.batch_gradient = results.get('optimized_batch_gradient', execute_stage_config.batch_gradient)


        logger.filesystem.sync['hedra.optimize'].info(f'{optimizer.metadata_string} - Optimization complete')

        context = {}
        for stage in pipeline_stages.values():
            context.update({
                context_key: context_value for context_key, context_value in stage.context.items() if context_key not in stage.context.known_keys
            })

        return {
            'stage': execute_stage.name,
            'config': execute_stage_config,
            'params': results,
            'context': context
        }

    except BrokenPipeError:

        try:
            optimizer._event_loop.close()
            loop.close()
        except RuntimeError:
            pass

        return {}
    
    except Exception as e:
        raise e


from typing import Any, Tuple, Dict
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.engines.types.grpc import GRPCResult
from .http2_event import HTTP2Event


class GRPCEvent(HTTP2Event):

    def __init__(
        self, 
        stage: Any, 
        result: GRPCResult
    ) -> None:
        super().__init__(
            stage,
            result
        )


import traceback
import asyncio
import uuid
import inspect
from typing import Any, Coroutine, Tuple, Dict
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.reporting.tags import Tag


class BaseEvent:

    __slots__ = (
        'stage_name',
        'event_id',
        'action_id',
        'name',
        'shortname',
        'error',
        'type',
        'source',
        'checks',
        'tags',
        'stage',
        'timings',
        'time'
    )

    def __init__(
        self, 
        stage: Any, 
        result: BaseResult
    ) -> None:

        self.event_id = str(uuid.uuid4())
        self.action_id = result.action_id

        self.name = None
        self.shortname = result.name
        self.error = result.error
        self.timings = {}
        self.type = result.type
        self.source = result.source
        self.checks = []
        self.stage_name = stage.name

        self.time = self.timings.get('total', 0)
        self.checks = [
            check for check in stage.hooks[HookType.CHECK] if check.name in result.checks
        ]

        self.tags = [
            Tag(
                tag.get('name'),
                tag.get('value')
            ) for tag in result.tags
        ]
        self.stage = None

    async def check_result(self):
        if self.error is None:
            await asyncio.gather(*[
                self._check_results(check) for check in self.checks
            ])

    async def _check_results(self, check: Coroutine):
        try:
            await check.call(self)
        except AssertionError:              
            error_message = traceback.format_exc().split(
                '\n'
            )[-3].strip()

            self.error = f'Check - {error_message} - for action - {self.name} - failed.'

    @property
    def fields(self):
        return ['name', 'stage', 'time', 'succeeded']

    @property
    def values(self):
        return [self.name, self.stage, self.time, self.success]

    @property
    def record(self):
        return {
            'name': self.name,
            'stage': self.stage,
            'time': self.time,
            'succeeded': self.success
        }

    @property
    def success(self):
        return self.error is None

    def tags_to_dict(self):
        return {
            tag.name: tag.value for tag in self.tags
        }


import asyncio
import time
import uuid
import traceback
from typing import Awaitable, Dict, Set, Tuple
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.http2.pipe import HTTP2Pipe
from hedra.core.engines.types.http2.connection import HTTP2Connection
from hedra.core.engines.types.common.ssl import get_http2_ssl_context
from hedra.core.engines.types.common.concurrency import (
    Semaphore
)
from .pool import HTTP2Pool
from .action import HTTP2Action
from .result import HTTP2Result


HTTP2ResponseFuture = Awaitable[HTTP2Action]
HTTP2BatchResponseFuture = Awaitable[Tuple[Set[HTTP2ResponseFuture], Set[HTTP2ResponseFuture]]]


class MercuryHTTP2Client:

    __slots__ = (
        'session_id',
        'timeouts',
        '_hosts',
        'registered',
        'closed',
        'sem',
        'pool',
        'active',
        'waiter',
        'ssl_context'
    )

    def __init__(self, concurrency: int = 10**3, timeouts: Timeouts = Timeouts(), reset_connections: bool=False) -> None:

        self.session_id = str(uuid.uuid4())
        self.timeouts = timeouts

        self._hosts = {}
        self.registered: Dict[str, HTTP2Action] = {}
        self.closed = False
        
        self.sem = Semaphore(value=concurrency)
        self.pool: HTTP2Pool = HTTP2Pool(concurrency, self.timeouts, reset_connections=reset_connections)
        self.pool.create_pool()
        self.active = 0
        self.waiter = None

        self.ssl_context = get_http2_ssl_context()
        

    async def wait_for_active_threshold(self):
        if self.waiter is None:
            self.waiter = asyncio.get_event_loop().create_future()
            await self.waiter

    async def prepare(self, request: HTTP2Action) -> Awaitable[None]:
        try:
            request.ssl_context = self.ssl_context

            if self._hosts.get(request.url.hostname) is None:
                socket_configs = await request.url.lookup()

                for ip_addr, configs in socket_configs.items():
                        for config in configs:

                            connection = HTTP2Connection(
                                0, 
                                self.timeouts, 
                                1,
                                self.pool.reset_connections,
                                self.pool.pool_type
                            )

                            try:
                                
                                await connection.connect(
                                    request.url.hostname,
                                    ip_addr,
                                    request.url.port,
                                    config,
                                    ssl=self.ssl_context,
                                    timeout=self.timeouts.connect_timeout
                                )

                                request.url.socket_config = config
                                break

                            except Exception as e:
                                pass

                        if request.url.socket_config:
                            break

                if request.url.socket_config is None:
                        raise Exception('Err. - No socket found.')
                
                self._hosts[request.url.hostname] = request.url.ip_addr
                        
            else:
                request.url.ip_addr = self._hosts[request.url.hostname]

            if request.is_setup is False:
                request.setup()

            self.registered[request.name] = request

        except Exception as e:
            raise e

    def extend_pool(self, increased_capacity: int):
        self.pool.size += increased_capacity
        self.pool.create_pool()
    
        self.sem = Semaphore(self.pool.size)

    def shrink_pool(self, decrease_capacity: int):
        self.pool.size -= decrease_capacity
        self.pool.create_pool()

        self.sem = Semaphore(self.pool.size)

    async def execute_prepared_request(self, action: HTTP2Action) -> HTTP2ResponseFuture:
        response = HTTP2Result(action)
        response.wait_start = time.monotonic()
        self.active += 1
        
        async with self.sem:

            pipe = self.pool.pipes.pop()
            connection = self.pool.connections.pop()
        
            try:

                if action.hooks.listen:
                    event = asyncio.Event()
                    action.hooks.channel_events.append(event)
                    await event.wait()

                if action.hooks.before:
                    action = await action.hooks.before.call(action, response)
                    action.setup()

                response.start = time.monotonic()

                stream = await connection.connect(
                    action.url.hostname,
                    action.url.ip_addr,
                    action.url.port,
                    action.url.socket_config,
                    ssl=action.ssl_context,
                    timeout=self.timeouts.connect_timeout
                )

                stream.encoder = action.hpack_encoder
     
                response.connect_end = time.monotonic()

                pipe.send_request_headers(action, stream)
  
                if action.encoded_data is not None:
                    await pipe.submit_request_body(action, stream)

                response.write_end = time.monotonic()

                await asyncio.wait_for(
                    pipe.receive_response(response, stream), 
                    timeout=self.timeouts.total_timeout
                )

                response.complete = time.monotonic()

                if action.hooks.after:
                    action = await action.hooks.after.call(action, response)
                    action.setup()

                if action.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel.call(response, action.hooks.listeners)
                        ) for channel in action.hooks.channels
                    ])

                    for listener in action.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            listener.setup()
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()      

                self.pool.pipes.append(pipe)
                self.pool.connections.append(connection)
                
            except Exception as e:
                response.complete = time.monotonic()
                response._status = 400
                response.error = str(e)

                self.pool.reset()

            self.active -= 1
            if self.waiter and self.active <= self.pool.size:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None

            return response

    async def close(self):
        if self.closed is False:
            await self.pool.close()
            self.closed = True



import psutil
import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Awaitable, Any
from hedra.core.graphs.simple_context import SimpleContext
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.tools.filesystem import open
from .hook import Hook


class RestoreHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        key: str=None,
        restore_filepath: str=None
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            hook_type=HookType.RESTORE
        )

        self.context_key = key
        self.restore_path = restore_filepath
        self.executor = ThreadPoolExecutor(
            max_workers=psutil.cpu_count(logical=False)
        )
        self.loop = None

    async def call(self, context: SimpleContext) -> None:
        self.loop = asyncio.get_event_loop()
        return await self.loop.run_in_executor(
            self.executor,
            self._run,
            context
        )

    def _run(self, context: SimpleContext):
        import asyncio
        import uvloop
        uvloop.install()

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        return loop.run_until_complete(self._load(context))

    async def _load(self, context: SimpleContext):

        restore_file = await open(self.restore_path, 'r')
        file_data = await restore_file.read()
        data = await self._call(
            context[self.context_key],
            file_data
        )

        if self.context_key == 'results':

            for stage_name, stage_data in data.items():

                if isinstance(stage_data, ResultsSet) is False:
                    results_set = ResultsSet(stage_data)
                    results_set.load_results()

                    data[stage_name] = results_set

        context[self.context_key] = data

        await restore_file.close()

        return context



from typing import Dict, Union
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.http2.result import HTTP2Result
from .action import GraphQLHTTP2Action


class GraphQLHTTP2Result(HTTP2Result):

    def __init__(self, action: GraphQLHTTP2Action, error: Exception = None) -> None:
        super().__init__(action, error)
        self.type = RequestTypes.GRAPHQL_HTTP2

    @classmethod
    def from_dict(cls, results_dict: Dict[str, Union[int, float, str,]]):

        action = GraphQLHTTP2Action(
            results_dict.get('name'),
            results_dict.get('url'),
            method=results_dict.get('method'),
            user=results_dict.get('user'),
            tags=results_dict.get('tags'),
        )

        response = GraphQLHTTP2Result(action, error=results_dict.get('error'))
        

        response.headers.update(results_dict.get('headers', {}))
        response.data = results_dict.get('data')
        response.status = results_dict.get('status')
        response.reason = results_dict.get('reason')
        response.checks = results_dict.get('checks')
     
        response.wait_start = results_dict.get('wait_start')
        response.start = results_dict.get('start')
        response.connect_end = results_dict.get('connect_end')
        response.write_end = results_dict.get('write_end')
        response.complete = results_dict.get('complete')

        return response


from typing import List, Optional, Dict, Union
from pydantic import BaseModel, StrictStr, StrictInt, StrictFloat, validator


class ActionHookValidator(BaseModel):
    weight: StrictInt
    order: StrictInt
    metadata: Optional[Dict[str, Union[StrictStr, StrictInt, StrictFloat]]]
    checks: Optional[List[StrictStr]]
    notify: Optional[List[StrictStr]]
    listen: Optional[List[StrictStr]]

    class Config:
        arbitrary_types_allowed = True

    @validator('weight', 'order')
    def validate_weight_and_order(cls, val):
        assert val > 0, "Order and weight values must be greater than zero!"


class ActionValidator:

    def __init__(
        __pydantic_self__, 
        weight: Optional[int]=1, 
        order: Optional[int]=1, 
        metadata: Optional[Dict[str, Union[str, int, float]]]={}, 
        checks: Optional[List[str]]=[],
        notify: Optional[List[str]]=[],
        listen: Optional[List[str]]=[]
    ) -> None:
        ActionHookValidator(
            weight=weight,
            order=order,
            metadata=metadata,
            checks=checks,
            notify=notify,
            listen=listen
        )



from typing import Optional, Dict, Union
from pydantic import BaseModel, StrictStr, StrictInt, StrictFloat


class TeardownHookValidator(BaseModel):
    key: Optional[StrictStr]
    metadata: Dict[StrictStr, Union[StrictInt, StrictStr, StrictFloat]]={}

    class Config:
        arbitrary_types_allowed = True


class TeardownValidator:

    def __init__(__pydantic_self__, key: str=None, metadata: Optional[Dict[str, Union[int, str, float]]]={}) -> None:
        TeardownHookValidator(
            key=key,
            metadata=metadata
        )


import json
import threading
import os
import time
import signal
import dill
import traceback
from hedra.core.graphs.hooks.registry.registrar import registrar
from typing import Dict, Any, List, Union, Tuple
from hedra.core.engines.client.config import Config
from hedra.core.graphs.hooks.registry.registry_types import ActionHook, TaskHook
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.engines.types.playwright import MercuryPlaywrightClient, ContextConfig
from hedra.core.engines.types.registry import RequestTypes
from hedra.core.engines.types.registry import registered_engines
from hedra.core.personas.persona_registry import registered_personas
from hedra.core.graphs.events.event import Event, EventHook
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.engine.engine_plugin import EnginePlugin
from hedra.plugins.types.persona.persona_plugin import PersonaPlugin
from hedra.core.graphs.stages.setup.setup import Setup
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.base.import_tools import (
    import_stages, 
    import_plugins, 
    set_stage_hooks
)
from hedra.core.personas import get_persona
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)

from hedra.core.graphs.stages.base.parallel.partition_method import PartitionMethod


async def start_execution(parallel_config: Dict[str, Any]):

    hedra_config_filepath = os.path.join(
        os.getcwd(),
        '.hedra.json'
    )

    hedra_config = {}
    if os.path.exists(hedra_config_filepath):
        with open(hedra_config_filepath, 'r') as hedra_config_file:
            hedra_config = json.load(hedra_config_file)

    logging_config = hedra_config.get('logging', {})
    logfiles_directory = logging_config.get(
        'logfiles_directory',
        os.getcwd()
    )

    log_level = logging_config.get('log_level', 'info')

    logging_manager.disable(
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.DISTRIBUTED_FILESYSTEM
    )

    logging_manager.update_log_level(log_level)
    logging_manager.logfiles_directory = logfiles_directory


    logger = HedraLogger()
    logger.initialize()
    await logger.filesystem.aio.create_logfile('hedra.core.log')
    logger.filesystem.create_filelogger('hedra.core.log')

    start = time.monotonic()

    graph_name = parallel_config.get('graph_name')
    graph_path: str= parallel_config.get('graph_path') 
    graph_id = parallel_config.get('graph_id')
    source_stage_name = parallel_config.get('source_stage_name')
    source_stage_context: Dict[str, Any] = parallel_config.get('source_stage_context')
    source_stage_linked_events: Dict[Tuple[HookType, str], List[Tuple[str, str]]] = parallel_config.get('source_stage_linked_events')
    source_stage_plugins = parallel_config.get('source_stage_plugins')
    source_stage_config = parallel_config.get('source_stage_config')
    source_stage_id = parallel_config.get('source_stage_id')
    thread_id = threading.current_thread().ident
    process_id = os.getpid()

    metadata_string = f'Graph - {graph_name}:{graph_id} - thread:{thread_id} - process:{process_id} - Stage: {source_stage_name}:{source_stage_id} - '

    execution_hooks = parallel_config.get('hooks')
    partition_method = parallel_config.get('partition_method')
    persona_config: Config = parallel_config.get('config')
    workers = parallel_config.get('workers')
    worker_id = parallel_config.get('worker_id')

    discovered: Dict[str, Stage] = import_stages(graph_path)
    plugins_by_type = import_plugins(graph_path)
    execute_stage: Stage = discovered.get(source_stage_name)()
    execute_stage: Stage = set_stage_hooks(execute_stage)

    execute_stage.context.update(source_stage_context)

    execution_hooks_count = len(execution_hooks)

    if partition_method == PartitionMethod.BATCHES and persona_config.optimized is False:
        if workers == worker_id:
            persona_config.batch_size = int(persona_config.batch_size/workers) + (persona_config.batch_size%workers)
        
        else:
            persona_config.batch_size = int(persona_config.batch_size/workers)


    stage_persona_plugins: List[str] = source_stage_plugins[PluginType.PERSONA]
    persona_plugins: Dict[str, PersonaPlugin] = plugins_by_type[PluginType.PERSONA]

    stage_engine_plugins: List[str] = source_stage_plugins[PluginType.ENGINE]
    engine_plugins: Dict[str, EnginePlugin] = plugins_by_type[PluginType.ENGINE]
    
    for plugin_name in stage_persona_plugins:
        plugin = persona_plugins.get(plugin_name)
        plugin.name = plugin_name
        registered_personas[plugin_name] = lambda config: plugin(config)
    
    for plugin_name in stage_engine_plugins:
        plugin = engine_plugins.get(plugin_name)
        plugin.name = plugin_name
        registered_engines[plugin_name] = lambda config: plugin(config)

    persona = get_persona(persona_config)
    persona.workers = workers


    await logger.filesystem.aio['hedra.core'].info(
        f'{metadata_string} - Executing {execution_hooks_count} actions with a batch size of {persona_config.batch_size} for {persona_config.total_time} seconds using Persona - {persona.type.capitalize()}'
    )

    setup_stage = Setup()
    setup_stage.plugins_by_type = plugins_by_type
    setup_stage.generation_setup_candidates = 1
    setup_stage.stages[execute_stage.name] = execute_stage
    setup_stage.config = source_stage_config

    stages: Dict[str, Stage] = await setup_stage.run()
    setup_execute_stage: Stage = stages.get(source_stage_name)

    actions = {
        hook.name: hook for hook in setup_execute_stage.hooks[HookType.ACTION]
    }

    actions.update({
        hook.name: hook for hook in setup_execute_stage.hooks[HookType.TASK]
    })

    actions_and_tasks: List[Union[ActionHook, TaskHook]] = [
        *setup_execute_stage.hooks.get(HookType.ACTION, []),
        *setup_execute_stage.hooks.get(HookType.TASK, [])
    ]


    for hook in actions_and_tasks:
        if hook.action.hooks.notify:
            for idx, listener_name in enumerate(hook.action.hooks.listeners):
                hook.action.hooks.listeners[idx] = actions.get(listener_name)


        if hook.action.type == RequestTypes.PLAYWRIGHT and isinstance(hook.session, MercuryPlaywrightClient):

            await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Setting up Playwright Session')

            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Browser Type: {persona_config.browser_type}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Device Type: {persona_config.device_type}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Locale: {persona_config.locale}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - geolocation: {persona_config.geolocation}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Permissions: {persona_config.permissions}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Color Scheme: {persona_config.color_scheme}')

            await hook.session.setup(ContextConfig(
                browser_type=persona_config.browser_type,
                device_type=persona_config.device_type,
                locale=persona_config.locale,
                geolocation=persona_config.geolocation,
                permissions=persona_config.permissions,
                color_scheme=persona_config.color_scheme,
                options=persona_config.playwright_options
            ))
    
    loaded_stages: Dict[str, Stage] = {
        setup_execute_stage.name: setup_execute_stage
    }
    pipeline_stages = {
        setup_execute_stage.name: setup_execute_stage
    }

    for event_target in source_stage_linked_events:
        target_hook_stage, target_hook_type, target_hook_name = event_target
        for event_source in source_stage_linked_events[event_target]:
            event_source_stage, _, event_hook_name = event_source
            
            if target_hook_stage == source_stage_name:

                source_stage: Stage = loaded_stages.get(event_source_stage)

                if source_stage is None:
                    source_stage = discovered.get(event_source_stage)()
                    source_stage = set_stage_hooks(source_stage)
                    
                    loaded_stages[source_stage.name] = source_stage


                source_stage = set_stage_hooks(source_stage)

                pipeline_stages[source_stage.name] = source_stage

                event_hooks = [hook.name for hook in source_stage.hooks[HookType.EVENT]]
                event_hook_idx = event_hooks.index(event_hook_name)

                event_hook: EventHook = source_stage.hooks[HookType.EVENT][event_hook_idx]

                target_hook_names = [hook.name for hook in setup_execute_stage.hooks[target_hook_type]]
                target_hook_idx = target_hook_names.index(target_hook_name)

                target_hook: Hook = setup_execute_stage.hooks[target_hook_type][target_hook_idx]
                target_hook.stage = source_stage_name
                target_hook.stage_instance = setup_execute_stage
                target_hook.stage_instance.hooks = setup_execute_stage.hooks

                event = Event(target_hook, event_hook)
                event.target_key = event_hook.key

                if target_hook_idx >= 0 and isinstance(target_hook, Event):
                    if event_hook.pre is True:
                        target_hook.pre_sources[event_hook.name] = event_hook
                        target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = target_hook

                    else:
                        target_hook.post_sources[event_hook.name] = event_hook
                        target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = target_hook
                    
                    registrar.all[event.name] = target_hook

                elif target_hook_idx >= 0:
                    target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = event
                    registrar.all[event.name] = event
                
                target_hook.stage_instance.linked_events[(target_hook.stage, target_hook.hook_type, target_hook.name)].append(
                    (event_hook.stage, event_hook.hook_type, event_hook.name)
                )

                setup_execute_stage.hooks[target_hook_type][target_hook_idx] = event
                
    persona.setup(setup_execute_stage.hooks, metadata_string)

    await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Starting execution')

    results = await persona.execute()

    elapsed = time.monotonic() - start

    await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Execution complete - Time (including addtional setup) took: {round(elapsed, 2)} seconds')

    for result in results:
        result.checks = [check.name for check in result.checks]

    context = {}
    for stage in pipeline_stages.values():
        context.update({
            context_key: context_value for context_key, context_value in stage.context.items() if context_key not in stage.context.known_keys
        })   
    
    return {
        'results': results,
        'total_results': len(results),
        'total_elapsed': persona.total_elapsed,
        'context': context
    }


def execute_actions(parallel_config: str):
    import asyncio
    import uvloop
    uvloop.install()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    def handle_loop_stop(signame):
        try:
            loop.close()

        except BrokenPipeError:
            os._exit(1)
                
        except RuntimeError:
            os._exit(1)

    for signame in ('SIGINT', 'SIGTERM'):
        loop.add_signal_handler(
            getattr(signal, signame),
            lambda signame=signame: handle_loop_stop(signame)
        )

    try:
        
        parallel_config: Dict[str, Any] = dill.loads(parallel_config)

        result = loop.run_until_complete(
            start_execution(parallel_config)
        )

        return result

    except BrokenPipeError:
        
        try:
            loop.close()
        except RuntimeError:
            pass

        return {}

    except Exception as e:
        raise e


from typing import Any, Tuple, Dict
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.engines.types.graphql import GraphQLResult
from .http_event import HTTPEvent


class GraphQLEvent(HTTPEvent):

    def __init__(
        self, 
        stage: Any, 
        result: GraphQLResult
    ) -> None:
        super(GraphQLEvent, self).__init__(
            stage,
            result
        )


from typing import Tuple, Optional
from pydantic import BaseModel, validator, StrictStr, StrictBool


class EventHookValidator(BaseModel):
    names: Optional[Tuple[StrictStr, ...]]
    pre: StrictBool=False
    key: Optional[StrictStr]

    @validator('names')
    def validate_names(cls, vals):
        assert len(vals) == len(set(vals))


class EventValidator:

    def __init__(__pydantic_self__, *names: Tuple[str, ...], pre: bool=False, key: Optional[str]=None) -> None:
        EventHookValidator(
            names=names,
            pre=pre,
            key=key
        )



import psutil
import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Type, Callable, Awaitable, Any, Dict
from hedra.core.graphs.simple_context import SimpleContext
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.tools.filesystem import open
from .hook import Hook


class SaveHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        key: str=None,
        checkpoint_filepath: str=None
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            hook_type=HookType.SAVE
        )

        self.context_key = key
        self.save_path = checkpoint_filepath
        self.executor = ThreadPoolExecutor(
            max_workers=psutil.cpu_count(logical=False)
        )
        self.loop = None

    async def call(self, context: SimpleContext) -> None:
        self.loop = asyncio.get_event_loop()
        return await self.loop.run_in_executor(
            self.executor,
            self._run,
            context
        )

    def _run(self, context: SimpleContext):
        import asyncio
        import uvloop
        uvloop.install()

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        return loop.run_until_complete(self._write(context))

    async def _write(self, context: SimpleContext) -> None:

        save_file = await open(self.save_path, 'w')
        context_data = context.get(self.context_key)

        if context_data and self.context_key == 'results':
            results_data: Dict[str, ResultsSet] = context_data
            for stage_name in context_data:
                context_data[stage_name] = results_data[stage_name].to_serializable()

        await save_file.write(
            await self._call(context_data)
        )

        await save_file.close()


        if context.get(self.context_key):
            context[self.context_key] = type(context[self.context_key])()


        




from typing import Any, Callable, Awaitable, Optional
from hedra.core.graphs.simple_context import SimpleContext
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from .hook import Hook


class ContextHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        store_key: str,
        load_key: Optional[str]=None
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            hook_type=HookType.CONTEXT
        )
        
        self.store_key = store_key
        self.load_key = load_key

    async def call(self, context: SimpleContext) -> None:

        if self.load_key:
            context[self.store_key] = await self.call(
                context[self.load_key]
            )

        else:
            context[self.store_key] = await self._call()

        return context


import asyncio
import networkx
import inspect
import threading
import os
from typing import List, Dict, Union, Any, Tuple, Coroutine
from collections import defaultdict
from hedra.core.graphs.events import Event
from hedra.core.graphs.stages.base.import_tools import set_stage_hooks
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.error import Error
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.graphs.hooks.registry.registry_types import EventHook
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook, HookType
from hedra.core.graphs.stages.base.parallel.batch_executor import BatchExecutor
from hedra.core.graphs.transitions.exceptions.exceptions import IsolatedStageError
from hedra.logging import HedraLogger
from hedra.plugins.types.engine.engine_plugin import EnginePlugin
from hedra.plugins.types.reporter.reporter_plugin import ReporterPlugin
from hedra.plugins.types.plugin_types import PluginType
from .transition import Transition
from .common import (
    invalid_transition
)
from .idle import (
    invalid_idle_transition
)


class TransitionAssembler:

    def __init__(
        self, 
        transition_types, 
        graph_name: str=None,
        graph_path: str=None,
        graph_id: str=None,
        cpus: int=None, 
        worker_id: int=None,
        core_config: Dict[str, Any]={}
    ) -> None:
        self.transition_types: Dict[Tuple[StageTypes, StageTypes], Coroutine] = transition_types
        self.core_config = core_config
        self.graph_name = graph_name
        self.graph_path = graph_path
        self.graph_id = graph_id
        self.generated_stages = {}
        self.transitions = {}
        self.instances_by_type: Dict[str, List[Stage]] = {}
        self.cpus = cpus
        self.worker_id = worker_id
        self.loop = asyncio.get_event_loop()
        self.hooks_by_type: Dict[HookType, Dict[str, Hook]] = defaultdict(dict)

        self.logging = HedraLogger()
        self.logging.initialize()

        self._thread_id = threading.current_thread().ident
        self._process_id = os.getpid()

        self._graph_metadata_log_string = f'Graph - {self.graph_name}:{self.graph_id} - thread:{self._thread_id} - process:{self._process_id} - '

    def generate_stages(self, stages: Dict[str, Stage]):

        stages_count = len(stages)
        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Generating - {stages_count} - stages')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Generating - {stages_count} - stages')

        self.instances_by_type = {}

        for stage in stages.values():
            self.instances_by_type[stage.stage_type] = []

        stage_types_count = len(self.instances_by_type)
        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Found - {stage_types_count} - unique stage types')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Found - {stage_types_count} - unique stage types')

        self.generated_stages: Dict[str, Stage] = {
            stage_name: stage() for stage_name, stage in stages.items()
        }

        for stage in self.generated_stages.values():
            stage.core_config = self.core_config
            stage.graph_name = self.graph_name
            stage.graph_path = self.graph_path
            stage.graph_id = self.graph_id

            stage.workers = self.cpus
            stage.worker_id = self.worker_id

            for hook_shortname, hook in registrar.reserved[stage.name].items():
                    hook._call = hook._call.__get__(stage, stage.__class__)
                    setattr(stage, hook_shortname, hook._call)


            stage = set_stage_hooks(stage)

            for hook_type in stage.hooks:
                self.hooks_by_type[hook_type].update({
                    hook.name: hook for hook in stage.hooks[hook_type]
                })

            methods = inspect.getmembers(stage, predicate=inspect.ismethod) 

            for _, method in methods:
                method_name = method.__qualname__
                hook: Hook = registrar.all.get(method_name)
                
                if hook:
                    hook._call = hook._call.__get__(stage, stage.__class__)
                    setattr(stage, hook.shortname, hook._call)

                    if inspect.ismethod(hook.call) is False:
                        hook.call = hook.call.__get__(stage, stage.__class__)
                        setattr(stage, hook.shortname, hook.call)

                    hook.stage = stage.name
                    hook.stage_instance: Stage = stage
                    
                    self.hooks_by_type[hook.hook_type][hook.name] = hook
                    stage.hooks[hook.hook_type].append(hook)
            
            self.instances_by_type[stage.stage_type].append(stage)


            event_hooks = []
            stage_event_hooks: List[EventHook] = stage.hooks[HookType.EVENT]
            for event_hook in stage_event_hooks:
                if event_hook.names is None or len(event_hook.names)  == 0:
                    event_hooks.append(event_hook)

            stage.hooks[HookType.EVENT] = event_hooks

        event_hooks: List[EventHook] = list(self.hooks_by_type.get(HookType.EVENT, {}).values())
        for event_hook in event_hooks:
            for target_hook_name in event_hook.names:    
                target_hook = registrar.all.get(target_hook_name)
                
                if target_hook:
                    self.logging.filesystem.sync['hedra.core'].info(
                        f'{self._graph_metadata_log_string} - Appendng Event - {event_hook.name}:{event_hook.hook_id} - to target Stage - {target_hook.stage}:{target_hook.stage_instance.stage_id} Event Hooks'
                    )

                    event = Event(target_hook, event_hook)
                    event.target_key = event_hook.key

                    target_hook_idx = -1

                    try:
                        target_hook_idx = target_hook.stage_instance.hooks[target_hook.hook_type].index(target_hook)

                    except ValueError:
                        pass

                    if target_hook_idx >= 0 and isinstance(target_hook, Event):
                        if event_hook.pre is True:
                            target_hook.pre_sources[event_hook.name] = event_hook
                            target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = target_hook

                        else:
                            target_hook.post_sources[event_hook.name] = event_hook
                            target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = target_hook
                        
                        registrar.all[event.name] = target_hook

                    elif target_hook_idx >= 0:
                        target_hook.stage_instance.hooks[target_hook.hook_type][target_hook_idx] = event
                        registrar.all[event.name] = event
                    
                    target_hook.stage_instance.linked_events[(target_hook.stage, target_hook.hook_type, target_hook.name)].append(
                        (event_hook.stage,  event_hook.hook_type, event_hook.name)
                    )
               
    
        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Successfully generated - {stages_count} - stages')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Successfully generated - {stages_count} - stages')

    def build_transitions_graph(self, topological_generations: List[List[str]], graph: networkx.Graph):

        self.logging.hedra.sync.debug(f'Buiding transitions matrix')
        self.logging.filesystem.sync['hedra.core'].debug(f'Buiding transitions matrix')

        transitions: List[List[Transition]] = []
        plugins: Dict[PluginType, Dict[str, Union[EnginePlugin, ReporterPlugin]]] = {
            PluginType.ENGINE: {},
            PluginType.OPTIMIZER: {},
            PluginType.PERSONA: {},
            PluginType.REPORTER: {}
        }

        for isolate_stage_name in networkx.isolates(graph):
            raise IsolatedStageError(
                self.generated_stages.get(isolate_stage_name)
            )
        
        for generation in topological_generations:

            generation_transitions = []

            stage_pool_size = self.cpus

            stages = {
                stage_name: self.generated_stages.get(stage_name) for stage_name in generation
            }
            parallel_stages = []

            no_workers_stages = [
                StageTypes.WAIT, 
                StageTypes.IDLE
            ]

            stages_count = len(stages)
            self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Provisioning workers - {stages_count} - stages')
            self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Provisioning workers- {stages_count} - stages')

            for stage in stages.values():

                for plugin_name, plugin in stage.plugins.items():
                    plugins[plugin.type][plugin_name] = plugin

                if stage.allow_parallel is False and stage.stage_type not in no_workers_stages:
                    stage.workers = 1
                    stage_pool_size -= 1

                    self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Stage - {stage.name} - provisioned - {stage.workers} - workers')
                    self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Stage - {stage.name} - provisioned - {stage.workers} - workers')

                    stage.executor = BatchExecutor(stage.workers)

                else:
                    parallel_stages.append((
                        stage.name,
                        stage
                    ))
            
            if len(parallel_stages) > 0:
                batch_executor = BatchExecutor(max_workers=stage_pool_size)

                batched_stages: List[Tuple[str, Stage, int]] = batch_executor.partion_stage_batches(parallel_stages)
                
                for _, stage, assigned_workers_count in batched_stages:

                    self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Stage - {stage.name} - provisioned - {assigned_workers_count} - workers')
                    self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Stage - {stage.name} - provisioned - {assigned_workers_count} - workers')

                    stage.workers = assigned_workers_count
                    stage.executor = BatchExecutor(max_workers=assigned_workers_count)

                    stages[stage.name] = stage

            for stage in stages.values():

                stage.plugins_by_type = plugins

                neighbors = list(graph.neighbors(stage.name))

                neighbors_count = len(neighbors)
                self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Discovered - {neighbors_count} - neighboring stages for stage - {stage.name}')
                self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Discovered - {neighbors_count} - neighboring stages for stage - {stage.name}')
                
                for neighbor in neighbors:
                    neighbor_stage = self.generated_stages.get(neighbor)

                    transition_action = self.transition_types.get((
                        stage.stage_type,
                        neighbor_stage.stage_type
                    ))

                    if transition_action == invalid_transition or transition_action == invalid_idle_transition:
                        invalid_transition_error, _ = self.loop.run_until_complete(
                            transition_action(stage, neighbor_stage)
                        )

                        raise invalid_transition_error

                    self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Created transition from - {stage.name} - to - {neighbor_stage.name}')
                    self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Created transition from - {stage.name} - to - {neighbor_stage.name}')

                    transition = Transition(
                        transition_action,
                        stage,
                        neighbor_stage
                    )

                    generation_transitions.append(transition)

            if len(generation_transitions) > 0:
                transitions.append(generation_transitions)

        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Transition matrix assemmbly complete')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Transition matrix assemmbly complete')

        return transitions

    def map_to_setup_stages(self, graph: networkx.DiGraph):

        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Mapping stages to requisite Setup stages')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Mapping stages to requisite Setup stages')
        
        idle_stages = self.instances_by_type.get(StageTypes.IDLE)
        for idle_stage in idle_stages:
            idle_stage.context.stages = {}
            idle_stage.context.visited = []
            idle_stage.context.results = {}
            idle_stage.context.results_stages = []
            idle_stage.context.summaries = {}
            idle_stage.context.paths = {}
            idle_stage.context.path_lengths = {}
            
        idle_stage_name = idle_stage.__class__.__name__

        complete_stage = self.instances_by_type.get(StageTypes.COMPLETE)[0]

        for stage_type in StageTypes:

            idle_stage.context.stages[stage_type] = {}

            for stage in self.instances_by_type.get(stage_type, []):

                stage_name = stage.__class__.__name__

                has_path = networkx.has_path(
                    graph, 
                    idle_stage_name,
                    stage_name
                )

                if has_path:
                    idle_stage.context.stages[stage_type][stage_name] = stage
                    paths = networkx.all_shortest_paths(graph, stage_name, complete_stage.name)
                
                    stage_paths = []
                    for path in paths:
                        stage_paths.extend(path)
                    
                    idle_stage.context.paths[stage_name] = stage_paths

                    path_lengths = networkx.all_pairs_shortest_path_length(graph)

                    stage_path_lengths = {}
                    for path_stage_name, path_lengths_set in path_lengths:

                        del path_lengths_set[path_stage_name]
                        stage_path_lengths[path_stage_name] = path_lengths_set

                    idle_stage.context.path_lengths[stage_name] = stage_path_lengths.get(stage_name)

        for stage in self.generated_stages.values():
            for idle_stage in idle_stages:
                stage.context = idle_stage.context

        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Mapped stages to requisite Setup stages')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Mapped stages to requisite Setup stages')
    
    def create_error_transition(self, error: Exception):

        from_stage = error.from_stage
            
        error_transition = self.transition_types.get((
            from_stage.stage_type,
            StageTypes.ERROR
        ))

        error_stage = Error()
        error_stage.graph_name = self.graph_name
        error_stage.graph_id = self.graph_id
        error_stage.error = error

        return Transition(
            error_transition,
            from_stage,
            error_stage
        )